import numpy as np
import polars as pl

from polars_utils.covariance import covariance_matrix


def test_covariance_matrix():
    iris = pl.read_csv(
        "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/iris.csv"
    ).select(pl.col(float))

    missing_pandas = iris.to_pandas().cov(ddof=0).to_numpy()
    missing_mine = iris.pipe(covariance_matrix)

    assert np.allclose(missing_pandas, missing_mine), (
        "unweighted cov matrix not equal to pandas"
    )

    weighted_np = np.cov(
        iris.to_numpy(),
        rowvar=False,
        aweights=iris.get_column("sepal_length").to_numpy(),
        ddof=0,
    )

    weighted_mine = iris.pipe(covariance_matrix, w="sepal_length")

    assert np.allclose(weighted_np, weighted_mine), (
        "weighted cov matrix not equal to numpy"
    )



from typing import Optional
import polars as pl


from polars_utils import IntoExpr, into_expr
from polars_utils.stats import mean


def reliability(
    estimates: pl.Expr,
    *,
    variances: IntoExpr,
    w: Optional[IntoExpr] = None,
    mu: Optional[pl.Expr] = None,
):
    variances = into_expr(variances)

    # point around which we compute variance
    mu = estimates.pipe(mean, w=w) if mu is None else mu

    # calculate reliability
    signal_variance = ((estimates - mu).pow(2) - variances).pipe(mean, w=w)

    return (signal_variance / (variances + signal_variance)).alias("reliability")


def shrink(
    estimates: pl.Expr,
    *,
    variances: Optional[IntoExpr] = None,
    w: Optional[IntoExpr] = None,
    mu: Optional[pl.Expr] = None,  # mean to shrink to
    rho: Optional[pl.Expr] = None,  # reliability
) -> pl.Expr:
    """
    https://libgen.li/adsa82270d055f6ee991539ac0533036e0dO9ZBZS4G
    """

    # point we shrink to
    if mu is None:
        mu = estimates.pipe(mean, w=w)

    # reliability
    if rho is None:
        if variances is None:
            raise ValueError("Must pass either variances or reliabilities.")

        rho = reliability(estimates, variances=variances, w=w, mu=mu)

    # shrink towards mean based on reliability
    return estimates * rho + mu * (1 - rho)



from itertools import product
from typing import Collection, Literal, Optional

import polars as pl

import numpy as np

from polars_utils import IntoExpr
from polars_utils.stats import cov


def covariance_matrix(
    df: pl.DataFrame,
    w: Optional[IntoExpr] = None,
    columns: IntoExpr | Collection[str] = pl.all(),
    nulls: Literal["drop_obs", "pairwise_complete", "error"] = "drop_obs",
) -> np.ndarray:
    if not isinstance(columns, Collection):
        columns = pl.selectors.expand_selector(df, columns, strict=False)

    if nulls == "drop_obs":
        df = df.drop_nulls(columns)
    elif nulls == "error":
        assert df.select(
            pl.all_horizontal(pl.col(c).is_null().any() for c in columns)
        ).item(), "nulls in data frame"
    elif nulls == "pairwise_complete":
        pass
    else:
        raise ValueError

    return (
        df.select(
            cov(a, b, w=w).alias(f"{a}|{b}") for a, b in product(columns, columns)
        )
        .to_numpy()
        .reshape((len(columns), len(columns)))
    )



from typing import Any, Iterable, Literal, Optional
import polars as pl
import numpy as np
from rpy2.robjects import numpy2ri, Formula, packages, default_converter, NULL
from functools import cache

from tqdm import tqdm

@cache
def get_gam():
    return packages.importr("mgcv").gam


def smooth(
    y: pl.Expr | str,
    *xs: pl.Expr | str,
    weights: pl.Expr | str | None = None,
    progress=False,
    **kwargs,
) -> pl.Expr:
    if progress:
        bar = tqdm(desc = "smoothing...")
        bar.clear()

    def f(cols):
        if progress:
            bar.update() # type: ignore

        return _smooth(
            y=cols[0],
            xs=cols[1:-1],
            weights=cols[-1],
            **kwargs,
        )
    
    return pl.map_groups(
        [y, *xs, weights or pl.lit(None, dtype=pl.Null)],
        f,
        return_dtype=pl.Float64,
        returns_scalar=False,
    ).alias(y if isinstance(y, str) else y.meta.output_name())


def _smooth(
    y: pl.Series,
    xs: Iterable[pl.Series],
    *,
    weights: Optional[pl.Series] = None,
    k: Optional[tuple[int, ...]] = None,
    family: Literal["gaussian", "binomial"] | Any = "gaussian",
) -> pl.Series:
    formula = (
        "y ~ te("
        + f"{','.join(f'x{i}' for i, _ in enumerate(xs))}"  # covariates
        + (f", k=c({','.join(map(str, k))})" if k else "")  # knots
        + ")"
    )

    with (numpy2ri.converter + default_converter).context():
        f = Formula(formula)

        f.environment["y"] = y.to_numpy()

        for i, x in enumerate(xs):
            f.environment[f"x{i}"] = x.to_numpy()

        gam = get_gam()
        model = gam(
            f,
            family=family,
            weights=NULL
            if weights is None or (weights.len() == 1 and weights.item() is None)
            else weights.to_numpy(),
        )
        fitted_values = np.asarray(model.rx2("fitted.values"))

    return pl.Series(y.name, fitted_values, nan_to_null=True, dtype=pl.Float64)
